package org.perl6.nqp.sixmodel.reprs;

import org.perl6.nqp.sixmodel.SixModelObject;
import org.perl6.nqp.sixmodel.StorageSpec;

public class VMArrayREPRData {
    public StorageSpec ss;
    public SixModelObject type;
}
