package org.perl6.nqp.sixmodel.reprs;

import org.perl6.nqp.sixmodel.SerializationContext;
import org.perl6.nqp.sixmodel.SixModelObject;

public class SCRefInstance extends SixModelObject {
    public SerializationContext referencedSC;
}
